use std::sync::Arc;
use std::marker::PhantomData;

use futures::StreamExt;
use pulsar::{
    consumer::{Consumer, ConsumerOptions, Message},
    producer::{Producer, ProducerOptions},
    Pulsar, TokioExecutor,
    DeserializeMessage, SerializeMessage,
};
use serde::{de::DeserializeOwned, Serialize};
use tokio::sync::Mutex;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum PulsarError {
    #[error("Failed to connect to Pulsar: {0}")]
    ConnectionError(String),
    #[error("Failed to create producer: {0}")]
    ProducerError(String),
    #[error("Failed to create consumer: {0}")]
    ConsumerError(String),
    #[error("Failed to send message: {0}")]
    SendError(String),
    #[error("Failed to receive message: {0}")]
    ReceiveError(String),
    #[error("Failed to deserialize message: {0}")]
    DeserializeError(String),
    #[error("Failed to serialize message: {0}")]
    SerializeError(String),
}

/// A type-agnostic Pulsar client that can work with any serializable type
pub struct PulsarClient {
    client: Arc<Pulsar<TokioExecutor>>,
}

impl PulsarClient {
    /// Creates a new Pulsar client
    pub async fn new(pulsar_url: &str) -> Result<Self, PulsarError> {
        let client = Pulsar::builder(pulsar_url, TokioExecutor)
            .build()
            .await
            .map_err(|e| PulsarError::ConnectionError(e.to_string()))?;

        Ok(Self {
            client: Arc::new(client),
        })
    }

    /// Creates a producer for a specific topic
    pub async fn create_producer<T: Serialize>(
        &self,
        topic: &str,
    ) -> Result<PulsarProducer<T>, PulsarError> {
        let producer = self
            .client
            .producer()
            .with_topic(topic)
            .with_options(ProducerOptions {
                ..Default::default()
            })
            .build()
            .await
            .map_err(|e| PulsarError::ProducerError(e.to_string()))?;

        Ok(PulsarProducer {
            producer: Arc::new(Mutex::new(producer)),
            _phantom: PhantomData,
        })
    }

    /// Creates a consumer for a specific topic
    pub async fn create_consumer<T: DeserializeOwned>(
        &self,
        topic: &str,
        subscription_name: &str,
    ) -> Result<PulsarConsumer<T>, PulsarError> {
        let consumer = self
            .client
            .consumer()
            .with_topic(topic)
            .with_subscription(subscription_name)
            .with_options(ConsumerOptions {
                ..Default::default()
            })
            .build::<Vec<u8>>()
            .await
            .map_err(|e| PulsarError::ConsumerError(e.to_string()))?;

        Ok(PulsarConsumer {
            consumer: Arc::new(Mutex::new(consumer)),
            _phantom: PhantomData,
        })
    }
}

/// A type-agnostic Pulsar producer
pub struct PulsarProducer<T> {
    producer: Arc<Mutex<Producer<TokioExecutor>>>,
    _phantom: PhantomData<T>,
}

impl<T: Serialize> PulsarProducer<T> {
    /// Sends a message to the topic
    pub async fn send(&self, message: &T) -> Result<(), PulsarError> {
        let serialized = serde_json::to_vec(message)
            .map_err(|e| PulsarError::SerializeError(e.to_string()))?;

        let mut producer = self.producer.lock().await;
        producer
            .send_non_blocking(serialized)
            .await
            .map_err(|e| PulsarError::SendError(e.to_string()))?;

        Ok(())
    }
}

/// A type-agnostic Pulsar consumer
pub struct PulsarConsumer<T> {
    consumer: Arc<Mutex<Consumer<Vec<u8>, TokioExecutor>>>,
    _phantom: PhantomData<T>,
}

impl<T: DeserializeOwned> PulsarConsumer<T> {
    /// Receives a message from the topic
    pub async fn receive(&self) -> Result<T, PulsarError> {
        let mut consumer = self.consumer.lock().await;
        let message = consumer
            .next()
            .await
            .ok_or_else(|| PulsarError::ReceiveError("No message available".to_string()))?
            .map_err(|e| PulsarError::ReceiveError(e.to_string()))?;

        let data = message.payload.data.as_slice();
        let deserialized = serde_json::from_slice(data)
            .map_err(|e| PulsarError::DeserializeError(e.to_string()))?;

        Ok(deserialized)
    }

    /// Acknowledges a message
    pub async fn acknowledge(&self, message: &Message<Vec<u8>>) -> Result<(), PulsarError> {
        let mut consumer = self.consumer.lock().await;
        consumer
            .ack(message)
            .await
            .map_err(|e| PulsarError::ReceiveError(e.to_string()))?;

        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use serde::{Deserialize, Serialize};

    #[derive(Debug, Serialize, Deserialize, PartialEq)]
    struct TestMessage {
        content: String,
    }

    #[tokio::test]
    async fn test_pulsar_client() {
        let client = PulsarClient::new("pulsar://localhost:6650")
            .await
            .expect("Failed to create Pulsar client");

        let topic = "non-persistent";
        let subscription = "test-subscription";

        let producer = client
            .create_producer::<TestMessage>(topic)
            .await
            .expect("Failed to create producer");

        let consumer = client
            .create_consumer::<TestMessage>(topic, subscription)
            .await
            .expect("Failed to create consumer");

        let test_message = TestMessage {
            content: "Hello, Pulsar!".to_string(),
        };

        producer
            .send(&test_message)
            .await
            .expect("Failed to send message");

        let received: TestMessage = consumer
            .receive()
            .await
            .expect("Failed to receive message");

        assert_eq!(received, test_message);
    }
}